﻿namespace HeuristicLab.Algorithms.PushGP.Expressions
{
    using HeuristicLab.Algorithms.PushGP.Interpreter;
    using HeuristicLab.Algorithms.PushGP.Stack;

    /// <summary>
    ///     Pushes a copy of an indexed item "deep" in the stack onto the top of the stack, without removing the deep item.
    ///     The index is taken from the INTEGER stack, and the indexing is done after the index is removed.
    /// </summary>
    /// <typeparam name="T">Stacktype</typeparam>
    public abstract class YankDuplicateExpression<T> : StatelessExpression
    {
        protected YankDuplicateExpression(string stringRepresentation) : base(stringRepresentation)
        {
        }

        public void Eval(IStack<T> stack, IStack<long> integerStack)
        {
            T item;
            if (this.TryGetItem(stack, integerStack, out item)) stack.Push(item);
        }

        public void Eval(IStack<Expression> stack, IStack<long> integerStack)
        {
            Expression item;
            if (this.TryGetItem(stack, integerStack, out item)) stack.Push(item);
        }

        public void Eval(IStack<long> integerStack)
        {
            if ((integerStack.Count < 2) || (integerStack.Top > integerStack.Count - 2) ||
                (integerStack.Top < 0)) return;

            var index = (int) integerStack.Pop();
            var item = integerStack.ReverseElementAt(index);

            integerStack.Push(item);
        }

        private bool TryGetItem<E>(IStack<E> stack, IStack<long> integerStack, out E item)
        {
            if ((integerStack.Count == 0) || (integerStack.Top >= stack.Count) || (integerStack.Top < 0) ||
                (stack.Count < 2))
            {
                item = default(E);
                return false;
            }

            var index = (int) integerStack.Pop();
            item = stack.ReverseElementAt(index);

            return true;
        }
    }

    public class IntegerYankDuplicateExpression : YankDuplicateExpression<long>
    {
        public IntegerYankDuplicateExpression() : base("INTEGER.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.IntegerStack);
        }
    }

    public class FloatYankDuplicateExpression : YankDuplicateExpression<double>
    {
        public FloatYankDuplicateExpression() : base("FLOAT.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.FloatStack, interpreter.IntegerStack);
        }
    }

    public class BooleanYankDuplicateExpression : YankDuplicateExpression<bool>
    {
        public BooleanYankDuplicateExpression() : base("BOOLEAN.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.BooleanStack, interpreter.IntegerStack);
        }
    }

    public class NameYankDuplicateExpression : YankDuplicateExpression<string>
    {
        public NameYankDuplicateExpression() : base("NAME.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.NameStack, interpreter.IntegerStack);
        }
    }

    public class ExecYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        public ExecYankDuplicateExpression() : base("EXEC.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.ExecStack, interpreter.IntegerStack);
        }
    }

    public class CodeYankDuplicateExpression : YankDuplicateExpression<Expression>
    {
        public CodeYankDuplicateExpression() : base("CODE.YANKDUP")
        {
        }

        public override void Eval(IPushGpInterpreter interpreter)
        {
            this.Eval(interpreter.CodeStack, interpreter.IntegerStack);
        }
    }
}